@@ -29,6 +29,7 @@ from ._qpdf import (
     AccessMode,
     ObjectStreamMode,
     PdfError,
+    Rectangle,
     StreamDecodeLevel,
     StreamParser,
     Token,
@@ -69,23 +70,29 @@ def augments(cls_cpp: Type[Any]):
     wrapper classes and subclasses can become problematic if the call stack
     crosses the C++/Python boundary multiple times.
 
-    Any existing methods may be used, regardless of whether they defined
+    Any existing methods may be used, regardless of whether they are defined
     elsewhere in the support class or in the target class.
 
-    For data fields to work, including @property accessors, the target class must be
+    For data fields to work, the target class must be
     tagged ``py::dynamic_attr`` in pybind11.
 
     Strictly, the target class does not have to be C++ or derived from pybind11.
     This works on pure Python classes too.
 
     THIS DOES NOT work for class methods.
+
+    (Alternative ideas: https://github.com/pybind/pybind11/issues/1074)
     """
+    ATTR_WHITELIST = {'__repr__', '__enter__', '__exit__', '__hash__'}
 
     def class_augment(cls, cls_cpp=cls_cpp):
         for name, member in inspect.getmembers(cls):
-            if hasattr(cls_cpp, name):  # For PyPy: don't replace existing methods
+            # Don't replace existing methods except those in our whitelist
+            if hasattr(cls_cpp, name) and name not in ATTR_WHITELIST:
                 continue
             if inspect.isfunction(member):
+                if member.__qualname__.startswith('object.'):
+                    continue  # To avoid breaking PyPy
                 member.__qualname__ = member.__qualname__.replace(
                     cls.__name__, cls_cpp.__name__
                 )
@@ -93,11 +100,11 @@ def augments(cls_cpp: Type[Any]):
             elif inspect.isdatadescriptor(member):
                 setattr(cls_cpp, name, member)
 
-        def block_init(self):
+        def disable_init(self):
             # Prevent initialization of the support class
             raise NotImplementedError(self.__class__.__name__ + '.__init__')
 
-        cls.__init__ = block_init
+        cls.__init__ = disable_init
         return cls
 
     return class_augment
@@ -115,6 +122,7 @@ def _single_page_pdf(page) -> bytes:
 
 def _mudraw(buffer, fmt) -> bytes:
     """Use mupdf draw to rasterize the PDF in the memory buffer"""
+    # mudraw cannot read from stdin so NamedTemporaryFile is required
     with NamedTemporaryFile(suffix='.pdf') as tmp_in:
         tmp_in.write(buffer)
         tmp_in.seek(0)
@@ -126,14 +134,12 @@ def _mudraw(buffer, fmt) -> bytes:
             stderr=PIPE,
             check=True,
         )
-        if proc.stderr:
-            raise RuntimeError(proc.stderr.decode())
         return proc.stdout
 
 
 @augments(Object)
 class Extend_Object:
-    def _repr_mimebundle_(self, include, exclude, **kwargs):
+    def _repr_mimebundle_(self, include=None, exclude=None):
         """Present options to IPython for rich display of this object
 
         See https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
@@ -144,7 +150,7 @@ class Extend_Object:
             and Name.Type in self
             and self.Type == Name.Page
         ):
-            return Page(self)._repr_mimebundle_(include, exclude, **kwargs)
+            return Page(self)._repr_mimebundle_(include, exclude)
         return None
 
     def _ipython_key_completions_(self):
@@ -152,7 +158,7 @@ class Extend_Object:
             return self.keys()
         return None
 
-    def emplace(self, other: Object):
+    def emplace(self, other: Object, retain=(Name.Parent,)):
         """Copy all items from other without making a new object.
 
         Particularly when working with pages, it may be desirable to remove all
@@ -160,6 +166,9 @@ class Extend_Object:
         of it, in a way that preserves all links and references to the original
         page. (Or similarly, for other Dictionary objects in a PDF.)
 
+        Any Dictionary keys in the iterable *retain* are preserved. By default,
+        /Parent is retained.
+
         When a page is assigned (``pdf.pages[0] = new_page``), only the
         application knows if references to the original the original page are
         still valid. For example, a PDF optimizer might restructure a page
@@ -176,11 +185,22 @@ class Extend_Object:
             >>> pdf.pages[0].emplace(pdf.pages[1])
             >>> pdf.pages[0].objgen
             (16, 0)  # Same object
+
+        .. versionchanged:: 2.11.1
+            Added the *retain* argument.
         """
         if not self.same_owner_as(other):
             raise TypeError("Objects must have the same owner for emplace()")
-        del_keys = set(self.keys()) - set(other.keys())
-        for k in other.keys():
+
+        # .keys() returns strings, so make all strings
+        retain = set(str(k) for k in retain)
+        self_keys = set(self.keys())
+        other_keys = set(other.keys())
+
+        assert all(isinstance(k, str) for k in (retain | self_keys | other_keys))
+
+        del_keys = self_keys - other_keys - retain
+        for k in (k for k in other_keys if k not in retain):
             self[k] = other[k]  # pylint: disable=unsupported-assignment-operation
         for k in del_keys:
             del self[k]  # pylint: disable=unsupported-delete-operation
@@ -270,11 +290,14 @@ class Extend_Pdf:
     def root(self):
         """
         Deprecated alias for .Root, the /Root object of the PDF.
+
+        .. deprecated:: 2.0
+            Use ``.Root``.
         """
         warn("Pdf.root is deprecated; use Pdf.Root", category=DeprecationWarning)
         return self.Root
 
-    def _repr_mimebundle_(self, **_kwargs):
+    def _repr_mimebundle_(self, include=None, exclude=None):
         """
         Present options to IPython or Jupyter for rich display of this object
 
@@ -312,6 +335,9 @@ class Extend_Pdf:
         You can delete the document information dictionary by deleting this property,
         ``del pdf.docinfo``. Note that accessing the property after deleting it
         will re-create with a new, empty dictionary.
+
+        .. versionchanged: 2.4
+            Added support for ``del pdf.docinfo``.
         """
         if Name.Info not in self.trailer:
             self.trailer.Info = self.make_indirect(Dictionary())
@@ -418,14 +444,8 @@ class Extend_Pdf:
         """
         Create a new pikepdf.Stream object that is attached to this PDF.
 
-        Args:
-            data (bytes): Binary data for the stream object
-            d: Dictionary portion of the stream object
-            kwargs: Keyword arguments to define the dictionary. Do not set
-                /Filter or /Length here as pikepdf will manage these.
-
-        Example:
-            pdf.make_stream(b"Binary data here", Type=Name.XObject, Subtype=Name.Image)
+        See:
+            :meth:`pikepdf.Stream.__new__`
 
         """
         return Stream(self, data, d, **kwargs)
@@ -434,7 +454,7 @@ class Extend_Pdf:
         self, *, page_size: Tuple[Numeric, Numeric] = (612.0, 792.0)
     ) -> Object:
         """
-        Add a blank page to this PD. If pages already exist, the page will be added to
+        Add a blank page to this PDF. If pages already exist, the page will be added to
         the end. Pages may be reordered using ``Pdf.pages``.
 
         The caller may add content to the page by modifying its objects after creating
@@ -503,10 +523,7 @@ class Extend_Pdf:
             b"%%EOF\n"
         )
 
-        if self.filename:
-            description = "closed file: " + self.filename
-        else:
-            description = "closed object"
+        description = "closed file: " + self.filename
         self._process(description, EMPTY_PDF)
         if getattr(self, '_tmp_stream', None):
             self._tmp_stream.close()
@@ -560,19 +577,12 @@ class Extend_Pdf:
 
         problems: List[str] = []
 
-        try:
-            self._decode_all_streams_and_discard()
-        except PdfError as e:
-            problems.append(str(e))
+        self._decode_all_streams_and_discard()
 
         discarding_parser = DiscardingParser()
-
         for basic_page in self.pages:
             page = Page(basic_page)
-            try:
-                page.parse_contents(discarding_parser)
-            except PdfError as e:
-                problems.append(str(e))
+            page.parse_contents(discarding_parser)
 
         for warning in self.get_warnings():
             problems.append("WARNING: " + warning)
@@ -729,7 +739,7 @@ class Extend_Pdf:
 
             stream_decode_level: Specifies how
                 to encode stream objects. See documentation for
-                ``StreamDecodeLevel``.
+                :class:`pikepdf.StreamDecodeLevel`.
 
             recompress_flate: When disabled (the default), qpdf does not
                 uncompress and recompress streams compressed with the Flate
@@ -777,14 +787,17 @@ class Extend_Pdf:
 
             :meth:`pikepdf.Pdf.remove_unreferenced_resources` before saving
             may eliminate unnecessary resources from the output file if there
-            any objects (such as images) that are referenced in a page's
+            are any objects (such as images) that are referenced in a page's
             Resources dictionary but never called in the page's content stream.
 
         .. note::
 
-            pikepdf can read PDFs will incremental updates, but always
-            any coalesces incremental updates into a single non-incremental
+            pikepdf can read PDFs with incremental updates, but always
+            coalesces any incremental updates into a single non-incremental
             PDF file when saving.
+
+        .. versionchanged:: 2.7
+            Added *recompress_flate*.
         """
         if not filename_or_stream and self._original_filename:
             filename_or_stream = self._original_filename
@@ -828,11 +841,11 @@ class Extend_Pdf:
         lazily loaded. Despite this restriction, pikepdf does not try to use any OS
         services to obtain an exclusive lock on the file. Some applications may
         want to attempt this or copy the file to a temporary location before
-        editing. This behaviour change if *allow_overwriting_input* is set: the whole
+        editing. This behaviour changes if *allow_overwriting_input* is set: the whole
         file is then read and copied to memory, so that pikepdf can overwrite it
         when calling ``.save()``.
 
-        When this is function is called with a stream-like object, you must ensure
+        When this function is called with a stream-like object, you must ensure
         that the data it returns cannot be modified, or undefined behavior will
         occur.
 
@@ -857,7 +870,8 @@ class Extend_Pdf:
             >>> pdf = Pdf.open("test.pdf", password="rosebud")
 
         Args:
-            filename_or_stream: Filename of PDF to open.
+            filename_or_stream: Filename or Python readable and seekable file
+                stream of PDF to open.
             password: User or owner password to open an
                 encrypted PDF. If the type of this parameter is ``str``
                 it will be encoded as UTF-8. If the type is ``bytes`` it will
@@ -898,7 +912,26 @@ class Extend_Pdf:
             TypeError: If the type of ``filename_or_stream`` is not
                 usable.
             FileNotFoundError: If the file was not found.
+
+        Note:
+            When *filename_or_stream* is a stream and the stream is located on a
+            network, pikepdf assumes that the stream using buffering and read caches
+            to achieve reasonable performance. Streams that fetch data over a network
+            in response to every read or seek request, no matter how small, will
+            perform poorly. It may be easier to download a PDF from network to
+            temporary local storage (such as ``io.BytesIO``), manipulate it, and
+            then re-upload it.
         """
+        if isinstance(filename_or_stream, bytes) and filename_or_stream.startswith(
+            b'%PDF-'
+        ):
+            warn(
+                "It looks like you called with Pdf.open(data) with a bytes-like object "
+                "containing a PDF. This will probably fail because this function "
+                "expects a filename or opened file-like object. Instead, please use "
+                "Pdf.open(BytesIO(data))."
+            )
+
         tmp_stream, original_filename = None, False
         if allow_overwriting_input:
             try:
@@ -929,14 +962,6 @@ class Extend_Pdf:
 
 @augments(_ObjectMapping)
 class Extend_ObjectMapping:
-    def __contains__(self, key) -> bool:
-        try:
-            self[key]
-        except KeyError:
-            return False
-        else:
-            return True
-
     def get(self, key, default=None) -> Object:
         try:
             return self[key]
@@ -971,6 +996,7 @@ def check_is_box(obj) -> bool:
 class Extend_Page:
     @property
     def mediabox(self):
+        "This page's /MediaBox, in PDF units."
         return self._get_mediabox(True)
 
     @mediabox.setter
@@ -980,6 +1006,10 @@ class Extend_Page:
 
     @property
     def cropbox(self):
+        """This page's effective /CropBox, in PDF units.
+
+        If the /CropBox is not defined, the /MediaBox is returned.
+        """
         return self._get_cropbox(True)
 
     @cropbox.setter
@@ -989,6 +1019,11 @@ class Extend_Page:
 
     @property
     def trimbox(self):
+        """This page's effective /TrimBox, in PDF units.
+
+        If the /TrimBox is not defined, the /CropBox is returned (and if
+        /CropBox is not defined, /MediaBox is returned).
+        """
         return self._get_trimbox(True)
 
     @trimbox.setter
@@ -998,6 +1033,7 @@ class Extend_Page:
 
     @property
     def resources(self):
+        """Return this pages resources dictionary."""
         return self.obj['/Resources']
 
     def add_resource(
@@ -1008,14 +1044,15 @@ class Extend_Page:
         *,
         prefix: str = '',
         replace_existing: bool = True,
-    ) -> str:
+    ) -> Name:
         """Adds a new resource to the page's Resources dictionary.
 
         If the Resources dictionaries do not exist, they will be created.
 
         Args:
             self: The object to add to the resources dictionary.
-            res: The resource dictionary object to add.
+            res: The dictionary object to insert into the resources
+                dictionary.
             res_type: Should be one of the following Resource dictionary types:
                 ExtGState, ColorSpace, Pattern, Shading, XObject, Font, Properties.
             name: The name of the object. If omitted, a random name will be
@@ -1029,7 +1066,14 @@ class Extend_Page:
             The name of the object.
 
         Example:
-            resource_name = Page(pdf.pages[0]).add_resource(formxobj, Name.XObject)
+            >>> resource_name = Page(pdf.pages[0]).add_resource(formxobj, Name.XObject)
+
+        .. versionadded:: 2.3
+
+        .. versionchanged:: 2.14
+            If *res* does not belong to the same `Pdf` that owns this page,
+            a copy of *res* is automatically created and added instead. In previous
+            versions, it was necessary to change for this case manually.
         """
         if not Name.Resources in self.obj:
             self.obj.Resources = Dictionary()
@@ -1054,13 +1098,78 @@ class Extend_Page:
                 else:
                     raise ValueError(f"Name {name} already exists in page /Resources")
 
-        resources[res_type][name] = res
+        resources[res_type][name] = res.with_same_owner_as(self.obj)
         return name
 
-    def __repr__(self):
-        return repr(self.obj).replace('Dictionary', 'Page')
+    def _over_underlay(
+        self, other, rect: Optional[Rectangle], under: bool = True
+    ) -> None:
+        formx = None
+        if isinstance(other, Page):
+            formx = other.as_form_xobject()
+        elif isinstance(other, Dictionary) and other.get(Name.Type) == Name.Page:
+            page = Page(other)
+            formx = page.as_form_xobject()
+        elif (
+            isinstance(other, Stream)
+            and other.get(Name.Type) == Name.XObject
+            and other.get(Name.Subtype) == Name.Form
+        ):
+            formx = other
 
-    def _repr_mimebundle_(self, include, exclude, **_kwargs):
+        if formx is None:
+            raise TypeError("other object is not something we can convert to FormX")
+
+        if rect is None:
+            rect = Rectangle(page.trimbox)
+
+        formx_placed_name = self.add_resource(formx, Name.XObject)
+        cs = self.calc_form_xobject_placement(formx, formx_placed_name, rect)
+
+        self.contents_add(cs, prepend=under)
+
+    def add_overlay(self, other: Union[Object, Page], rect: Optional[Rectangle] = None):
+        """Overlay another object on this page.
+
+        Overlays will be drawn after all previous content, potentially drawing on top
+        of existing content.
+
+        Args:
+            other: A Page or Form XObject to render as an overlay on top of this
+                page.
+            rect: The PDF rectangle (in PDF units) in which to draw the overlay.
+                If omitted, this page's trimbox, cropbox or mediabox will be used.
+
+        .. versionadded:: 2.14
+        """
+        return self._over_underlay(other, rect, under=False)
+
+    def add_underlay(
+        self, other: Union[Object, Page], rect: Optional[Rectangle] = None
+    ):
+        """Underlay another object beneath this page.
+
+        Underlays will be drawn before all other content, so they may be overdrawn
+        partially or completely.
+
+        Args:
+            other: A Page or Form XObject to render as an underlay underneath this
+                page.
+            rect: The PDF rectangle (in PDF units) in which to draw the underlay.
+                If omitted, this page's MediaBox will be used.
+
+        .. versionadded:: 2.14
+        """
+        return self._over_underlay(other, rect, under=True)
+
+    def __repr__(self):
+        return (
+            repr(self.obj)
+            .replace('Dictionary', 'Page', 1)
+            .replace('(Type="/Page")', '', 1)
+        )
+
+    def _repr_mimebundle_(self, include=None, exclude=None):
         data = {}
         bundle = {'application/pdf', 'image/png'}
         if include:
@@ -1082,3 +1191,12 @@ class Extend_Page:
 class Extend_Token:
     def __repr__(self):
         return f'pikepdf.Token({self.type_}, {self.raw_value})'
+
+
+@augments(Rectangle)
+class Extend_Rectangle:
+    def __repr__(self):
+        return f'pikepdf.Rectangle({self.llx}, {self.lly}, {self.urx}, {self.ury})'
+
+    def __hash__(self):
+        return hash((self.llx, self.lly, self.urx, self.ury))
